@@ -1,26 +1,29 @@
 from __future__ import annotations
+from urllib.parse import urljoin, urlparse
 from aiohttp import ClientSession
-from asyncio import gather as aio_gather
 from base64 import b64decode
 from math import ceil
 from pathlib import Path
-from pydantic import BaseSettings
 from typing import Any, Dict, List, Optional, Union
+
+import asyncio
 import requests
 import datetime
+import sys
+import warnings
 
 try:
-    # from ._utils import Decorators #, create_random_suffix
     from ._base import OAuthClient, AsyncOAuthClient
 except ImportError:
-    # from _utils import Decorators #, create_random_suffix
     from _base import OAuthClient, AsyncOAuthClient
 
 
-class Defaults(BaseSettings):
-    PAGE_SIZE: int = 50
-    TIME_OUT: int = 60
-    DATE_FROM: datetime.date = datetime.date.today()
+if (sys.platform.startswith('win')
+        and sys.version_info[0] == 3
+        and sys.version_info[1] >= 8):
+    policy = asyncio.WindowsSelectorEventLoopPolicy()
+    asyncio.set_event_loop_policy(policy)
+
 
 class JsonObject:
     """ Convert a dict['element'] for access like object.property. """
@@ -42,6 +45,10 @@ class Client(OAuthClient):
         @property
         def first_tenant_id(self):
             """ Return the id of the first available tenant. """
+            if not self.tenants or len(self.tenants) == 0:
+                raise ConnectionError(
+                    'The user does not have access to any tenant.'
+                )
             return getattr(self.tenants[0], 'Id', None)
 
         def get_tenant_id(self, filters_: dict):
@@ -51,7 +58,7 @@ class Client(OAuthClient):
             :param filters_: Dict with parameters to eval. Eg.
                 {"DBName": "Name_Of_Database", "TenantName": "Tenant_Test"}.
             :param tenant_name: (optional) Str with name of tenant.
-            :param \*\*kwargs: Optional arguments to filter.
+            :param **kwargs: Optional arguments to filter.
             :return: :class:`int` object
             :rtype: int
             """
@@ -106,7 +113,9 @@ class Client(OAuthClient):
         url: Union[str, Path],
         username: str,
         pwd: str,
+        admin_url: Optional[Union[str, Path]] = None,
         session: Optional[requests.Session] = None,
+        session_cfg: Optional[Dict[str, Any]] = None,
         tenant_filter: Optional[Dict[str, Any]] = None,
     ) -> None:
         """ Create a conection with visma app using recived parameters.
@@ -125,7 +134,8 @@ class Client(OAuthClient):
             url=url,
             username=username,
             pwd=pwd,
-            session=session
+            session=session,
+            session_cfg=session_cfg
         )
 
         # dict to filter tenant
@@ -141,6 +151,29 @@ class Client(OAuthClient):
         self.account = None
         self.tenant = None
 
+        # auth url
+        self.admin_url = admin_url
+        if admin_url:
+            self.client_admin_url = urlparse(admin_url) 
+        else:
+            # deprecation warning
+            warnings.warn(
+                message='You are using old Visma API. Update to new servers.',
+                category=DeprecationWarning
+            )
+
+            self.client_admin_url = urlparse(urljoin(
+                self.client_url.geturl(),
+                '/Admin/'
+            ))
+            self.client_url = urlparse(urljoin(
+                self.client_url.geturl(),
+                '/WebApi/')
+            )
+
+    def __enter__(self, *args, **kwargs) -> Client:
+        return super().__enter__(*args, **kwargs)
+
     def refresh_headers(
         self,
         auth_data: Optional[Dict[str, Any]] = None,
@@ -197,11 +230,18 @@ class Client(OAuthClient):
 
         return getattr(self.authentication, 'is_expired', None)
 
+    def get_account(self, path: str) -> Dict[str, Any]:
+        # data prepare
+        account_url = urljoin(self.client_admin_url.geturl(), 'account/')
+        path_url = urljoin(account_url, path)
+        
+        return self.get(url=path_url)
+
     def set_account(self) -> None:
         self.account = self.Account(
-            tenants=self.get(path='/Admin/account/tenants'),
-            user_info=self.get(path='/Admin/account/user-info'),
-            roles=self.get(path='/Admin/account/roles')
+            tenants=self.get_account(path='tenants'),
+            user_info=self.get_account(path='user-info'),
+            roles=self.get_account(path='roles')
         )
 
     def login(self) -> None:
@@ -213,6 +253,11 @@ class Client(OAuthClient):
         self.refresh_headers(remove_auth=True, remove_content_type=True)
 
         # data prepare
+        # data prepare
+        login_url = urljoin(
+            self.client_admin_url.geturl(),
+            'authentication/login'
+        )
         data = {
             "username": self.username,
             "password": b64decode(self.pwd).decode('utf-8'),
@@ -220,7 +265,7 @@ class Client(OAuthClient):
         }
 
         # consulting nettime
-        json_data = self.post(path='/Admin/authentication/login', data=data)
+        json_data = self.post(url=login_url, data=data)
 
         # self.authentication = self.Authentication(**json_data)
         self.refresh_headers(auth_data=json_data)
@@ -248,8 +293,14 @@ class Client(OAuthClient):
         if not self.is_connected:
             return
 
+        # data prepare
+        logout_url = urljoin(
+            self.client_admin_url.geturl(),
+            'authentication/logout'
+        )
+
         # disconnect ...
-        response = self.post(path='/Admin/authentication/logout')
+        response = self.post(url=logout_url)
 
         self.refresh_headers(remove_auth=True, remove_content_type=True)
         self.refresh_session()
@@ -287,7 +338,6 @@ class Client(OAuthClient):
         all_pages: bool = False,
         get_detail: bool = False,
         get_extensions: Optional[List[str]] = None,
-        bulk_get: bool = False,
         **kwargs
     ):
         """
@@ -316,7 +366,7 @@ class Client(OAuthClient):
         """
         
         # path prepare
-        path = '/WebApi/employees{}{}'.format(
+        path = 'employees{}{}'.format(
             f'/{employee}' if employee else '',
             f'/{extension}' if employee and extension else '',
         )
@@ -341,37 +391,14 @@ class Client(OAuthClient):
                 response.get('totalCount') / params.get('pageSize')
             )
             
-            # TODO: fix it
-            # only with bulk_get param
-            if bulk_get:
-                together = [{'page': i} for i in range(2, num_pages + 1)]
-
-                # get list of pages
-                pages_result = self.bulk_get(
-                    path=path,
-                    params=params,
-                    together=together,
-                    max_workers=kwargs.get('max_workers', 50)
+            # recursive get and extend response values
+            for i in range(2, num_pages + 1):
+                # update page
+                params['page'] = i
+                response['values'].extend(
+                    self.get(path=path, params=params).get('values')
                 )
 
-                # get values and add to structure
-                for pr in pages_result:
-                    response['values'].extend(pr.get('values'))
-
-            else:
-                # recursive get and extend response values
-                for i in range(2, num_pages + 1):
-                    # update page
-                    params['page'] = i
-                    response['values'].extend(
-                        self.get(path=path, params=params).get('values')
-                    )
-
-        
-        # put elements in dict if is ask extensions
-        if bulk_get and extension and kwargs.get('group_results', None):
-            return {employee: {extension: response}}
-
         if get_detail and not employee:
             response['values'] = [self.get_employee_detail(
                 employee=_employee.get('externalId'),
@@ -392,7 +419,7 @@ class Client(OAuthClient):
             raise KeyError("No se pueden especificar un address y extension.")
         
         # path prepare
-        path = '/WebApi/addresses{}{}'.format(
+        path = 'addresses{}{}'.format(
             f'/{address}' if address else '',
             f'/{extension}' if extension else '',
         )
@@ -410,7 +437,7 @@ class Client(OAuthClient):
     def get_birth_places(self, **kwargs):
         
         # path prepare
-        path = '/WebApi/birth-places'
+        path = 'birth-places'
 
         # parameters prepare
         params = {
@@ -427,7 +454,7 @@ class Client(OAuthClient):
     def get_countries(self, **kwargs):
 
         # path prepare
-        path = '/WebApi/countries'
+        path = 'countries'
 
         # parameters prepare
         params = {
@@ -443,7 +470,7 @@ class Client(OAuthClient):
     def get_family_members(self, **kwargs):
 
         # path prepare
-        path = '/WebApi/family-members/types'
+        path = 'family-members/types'
 
         # parameters prepare
         params = {
@@ -459,7 +486,7 @@ class Client(OAuthClient):
             **kwargs):
 
         # path prepare
-        path = '/WebApi/journals{}{}'.format(
+        path = 'journals{}{}'.format(
             f'/{journal}' if journal and extension else '',
             f'/{extension}' if journal and extension else '',
         )
@@ -486,7 +513,7 @@ class Client(OAuthClient):
     def get_leaves(self, extension: str = None, **kwargs):
 
         # path prepare
-        path = '/WebApi/leaves{}'.format(
+        path = 'leaves{}'.format(
             f'/{extension}' if extension else '',
         )
 
@@ -516,7 +543,7 @@ class Client(OAuthClient):
     def get_loans(self, **kwargs):
 
         # path prepare
-        path = '/WebApi/loans'
+        path = 'loans'
 
         # getting default date
         today = datetime.date.today()
@@ -537,7 +564,7 @@ class Client(OAuthClient):
     def get_nationalities(self, **kwargs):
 
         # path prepare
-        path = '/WebApi/loans'
+        path = 'loans'
 
         # parameters prepare
         params = {
@@ -553,7 +580,7 @@ class Client(OAuthClient):
     def get_pay_elements(self, employeeExternalId: str, **kwargs):
 
         # path prepare
-        path = '/WebApi/pay-elements/individual'
+        path = 'pay-elements/individual'
 
         # parameters prepare
         params = {
@@ -615,15 +642,15 @@ class Client(OAuthClient):
         """
 
         # path prepare
-        path = '/WebApi/pay-elements/individual'
+        path = 'pay-elements/individual'
 
         # request.get -> json
-        return self.post(path=path, json={"values": values}, **kwargs)
+        return self.post(path=path, json=values, **kwargs)
 
     def get_payments(self, extension: str, **kwargs):
 
         # path prepare
-        path = f'/WebApi/payments/{extension}'
+        path = f'payments/{extension}'
 
         # parameters prepare
         params = {
@@ -638,7 +665,7 @@ class Client(OAuthClient):
     def get_payrolls(self, extension: str, **kwargs):
 
         # path prepare
-        path = f'/WebApi/payrolls/{extension}'
+        path = f'payrolls/{extension}'
 
         # parameters prepare
         params = {
@@ -668,7 +695,7 @@ class Client(OAuthClient):
     def get_phases(self, phase: str = None, **kwargs):
 
         # path prepare
-        path = '/WebApi/phases{}'.format(
+        path = 'phases{}'.format(
             f'/{phase}' if phase else '',
         )
 
@@ -692,7 +719,7 @@ class Client(OAuthClient):
             raise KeyError("No se pueden especificar un phone y extension.")
 
         # path prepare
-        path = '/WebApi/phones{}{}'.format(
+        path = 'phones{}{}'.format(
             f'/{phone}' if phone else '',
             f'/{extension}' if extension else '',
         )
@@ -710,7 +737,7 @@ class Client(OAuthClient):
     def get_scales(self, scale: int, **kwargs):
 
         # path prepare
-        path = '/WebApi/scales'
+        path = 'scales'
 
         # parameters prepare
         params = {
@@ -728,7 +755,7 @@ class Client(OAuthClient):
     def get_seizures(self, startDate: str, **kwargs):
 
         # path prepare
-        path = '/WebApi/seizures'
+        path = 'seizures'
 
         # parameters prepare
         params = {
@@ -746,7 +773,7 @@ class Client(OAuthClient):
     def get_structures(self, extension: str = None, **kwargs):
 
         # path prepare
-        path = '/WebApi/structures{}'.format(
+        path = 'structures{}'.format(
             f'/{extension}' if extension else ''
         )
 
@@ -770,7 +797,7 @@ class Client(OAuthClient):
             raise KeyError("Debe especificar un applicationName.")
 
         # path prepare
-        path = '/WebApi/sync{}'.format(
+        path = 'sync{}'.format(
             f'/{extension}' if extension else ''
         )
 
@@ -793,7 +820,7 @@ class Client(OAuthClient):
     def get_time_management(self, extension: str = None, **kwargs):
         
         # path prepare
-        path = '/WebApi/sync{}'.format(
+        path = 'sync{}'.format(
             f'/{extension}' if extension else ''
         )
 
@@ -834,7 +861,7 @@ class Client(OAuthClient):
         """
 
         # request.get -> json
-        return self.get(path='/WebApi/version')
+        return self.get(path='version')
 
 
 class AsyncClient(AsyncOAuthClient):
@@ -850,6 +877,10 @@ class AsyncClient(AsyncOAuthClient):
         @property
         def first_tenant_id(self):
             """ Return the id of the first available tenant. """
+            if not self.tenants or len(self.tenants) == 0:
+                raise ConnectionError(
+                    'The user does not have access to any tenant.'
+                )
             return getattr(self.tenants[0], 'Id', None)
 
         def get_tenant_id(self, filters_: dict):
@@ -903,6 +934,7 @@ class AsyncClient(AsyncOAuthClient):
         url: Union[str, Path],
         username: str,
         pwd: str,
+        admin_url: Optional[Union[str, Path]] = None,
         session: Optional[ClientSession] = None,
         tenant_filter: Optional[Dict[str, Any]] = None,
     ) -> None:
@@ -938,9 +970,29 @@ class AsyncClient(AsyncOAuthClient):
         self.account = None
         self.tenant = None
 
-        # defaults
-        self.defaults = Defaults()
+        # auth url
+        self.admin_url = admin_url
+        if admin_url:
+            self.client_admin_url = urlparse(admin_url) 
+        else:
+            # deprecation warning
+            warnings.warn(
+                message='You are using old Visma API. Update to new servers.',
+                category=DeprecationWarning
+            )
 
+            self.client_admin_url = urlparse(urljoin(
+                self.client_url.geturl(),
+                '/Admin/'
+            ))
+            self.client_url = urlparse(urljoin(
+                self.client_url.geturl(),
+                '/WebApi/')
+            )
+
+    async def __aenter__(self) -> AsyncClient:
+        return await super().__aenter__()
+    
     def refresh_headers(
         self,
         auth_data: Optional[Dict[str, Any]] = None,
@@ -1015,11 +1067,19 @@ class AsyncClient(AsyncOAuthClient):
 
         return getattr(self.authentication, 'is_expired', None)
 
+    async def get_account(self, path: str) -> Dict[str, Any]:
+        # data prepare
+        account_url = urljoin(self.client_admin_url.geturl(), 'account/')
+        path_url = urljoin(account_url, path)
+        
+        return await self.get(url=path_url)
+
     async def set_account(self) -> None:
+        
         self.account = self.Account(
-            tenants=await self.get(path='/Admin/account/tenants'),
-            user_info=await self.get(path='/Admin/account/user-info'),
-            roles=await self.get(path='/Admin/account/roles')
+            tenants=await self.get_account(path='tenants'),
+            user_info=await self.get_account(path='user-info'),
+            roles=await self.get_account(path='roles')
         )
 
     async def login(self) -> None:
@@ -1031,17 +1091,17 @@ class AsyncClient(AsyncOAuthClient):
         self.refresh_headers(remove_auth=True, remove_content_type=True)
 
         # data prepare
+        login_url = urljoin(
+            self.client_admin_url.geturl(),
+            'authentication/login'
+        )
         data = {
             "username": self.username,
             "password": b64decode(self.pwd).decode('utf-8'),
             "grant_type": "password"
         }
-
-        # consulting nettime
-        json_data = await self.post(
-            path='/Admin/authentication/login',
-            data=data
-        )
+        # consulting visma
+        json_data = await self.post(url=login_url, data=data)
 
         # self.authentication = self.Authentication(**json_data)
         self.refresh_headers(auth_data=json_data)
@@ -1069,8 +1129,14 @@ class AsyncClient(AsyncOAuthClient):
         if not self.is_connected:
             return
 
+        # data prepare
+        logout_url = urljoin(
+            self.client_admin_url.geturl(),
+            'authentication/logout'
+        )
+
         # disconnect ...
-        response = await self.post(path='/Admin/authentication/logout')
+        response = await self.post(url=logout_url)
 
         self.refresh_headers(remove_auth=True, remove_content_type=True)
         self.refresh_session()
@@ -1096,7 +1162,7 @@ class AsyncClient(AsyncOAuthClient):
                 all_pages=True
             ) for extension in extensions]
             tasks = self.get_async_tasks(*coroutines)
-            extensions_detail = await aio_gather(*tasks)
+            extensions_detail = await asyncio.gather(*tasks)
 
             for ext_detail in extensions_detail:
                 ext_name = f'_{ext_detail.get("extension")}'
@@ -1116,7 +1182,7 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/employees{}{}'.format(
+        path = 'employees{}{}'.format(
             f'/{employee}' if employee else '',
             f'/{extension}' if employee and extension else '',
         )
@@ -1151,7 +1217,7 @@ class AsyncClient(AsyncOAuthClient):
                 params={**params, 'page': i}
             ) for i in range(2, num_pages + 1)]
             tasks = self.get_async_tasks(*coroutines)
-            responses = await aio_gather(*tasks)
+            responses = await asyncio.gather(*tasks)
 
             for child_response in responses:
                 response['values'].extend(child_response.get('values'))
@@ -1162,7 +1228,7 @@ class AsyncClient(AsyncOAuthClient):
                 extensions=get_extensions
             ) for _employee in response.get('values')]
             tasks = self.get_async_tasks(*coroutines)
-            response['values'] = await aio_gather(*tasks)
+            response['values'] = await asyncio.gather(*tasks)
             # response['values'] = responses
 
         if extension:
@@ -1184,7 +1250,7 @@ class AsyncClient(AsyncOAuthClient):
             raise KeyError("No se pueden especificar un address y extension.")
 
         # path prepare
-        path = '/WebApi/addresses{}{}'.format(
+        path = 'addresses{}{}'.format(
             f'/{address}' if address else '',
             f'/{extension}' if extension else '',
         )
@@ -1201,7 +1267,7 @@ class AsyncClient(AsyncOAuthClient):
     async def get_birth_places(self, **kwargs) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/birth-places'
+        path = 'birth-places'
 
         # parameters prepare
         params = {
@@ -1218,7 +1284,7 @@ class AsyncClient(AsyncOAuthClient):
     async def get_countries(self, **kwargs) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/countries'
+        path = 'countries'
 
         # parameters prepare
         params = {
@@ -1234,7 +1300,7 @@ class AsyncClient(AsyncOAuthClient):
     async def get_family_members(self, **kwargs) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/family-members/types'
+        path = 'family-members/types'
 
         # parameters prepare
         params = {
@@ -1255,7 +1321,7 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/journals{}{}'.format(
+        path = 'journals{}{}'.format(
             f'/{journal}' if journal and extension else '',
             f'/{extension}' if journal and extension else '',
         )
@@ -1287,7 +1353,7 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/leaves{}'.format(
+        path = 'leaves{}'.format(
             f'/{extension}' if extension else '',
         )
 
@@ -1316,7 +1382,7 @@ class AsyncClient(AsyncOAuthClient):
 
     async def get_loans(self, **kwargs) -> Union[Dict, List]:
         # path prepare
-        path = '/WebApi/loans'
+        path = 'loans'
 
         # parameters prepare
         params = {
@@ -1337,7 +1403,7 @@ class AsyncClient(AsyncOAuthClient):
     async def get_nationalities(self, **kwargs) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/loans'
+        path = 'loans'
 
         # parameters prepare
         params = {
@@ -1358,7 +1424,7 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/pay-elements/individual'
+        path = 'pay-elements/individual'
 
         # parameters prepare
         params = {
@@ -1375,6 +1441,7 @@ class AsyncClient(AsyncOAuthClient):
         # request.get -> json
         return await self.get(path=path, params=params)
 
+
     async def post_pay_elements(
         self,
         *,
@@ -1383,10 +1450,12 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/pay-elements/individual'
+        path = 'pay-elements/individual'
 
         # request.get -> json
-        return await self.post(path=path, json={"values": values}, **kwargs)
+        # return await self.post(path=path, json={"values": values}, **kwargs)
+        return await self.post(path=path, json=values, **kwargs)
+
 
     async def get_payments(
         self,
@@ -1396,7 +1465,7 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = f'/WebApi/payments/{extension}'
+        path = f'payments/{extension}'
 
         # parameters prepare
         params = {
@@ -1416,7 +1485,7 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = f'/WebApi/payrolls/{extension}'
+        path = f'payrolls/{extension}'
 
         # parameters prepare
         params = {
@@ -1451,7 +1520,7 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/phases{}'.format(
+        path = 'phases{}'.format(
             f'/{phase}' if phase else '',
         )
 
@@ -1481,7 +1550,7 @@ class AsyncClient(AsyncOAuthClient):
             raise KeyError("Cant use `phone` and `extension`.")
 
         # path prepare
-        path = '/WebApi/phones{}{}'.format(
+        path = 'phones{}{}'.format(
             f'/{phone}' if phone else '',
             f'/{extension}' if extension else '',
         )
@@ -1504,7 +1573,7 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/scales'
+        path = 'scales'
 
         # parameters prepare
         params = {
@@ -1527,7 +1596,7 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/seizures'
+        path = 'seizures'
 
         # parameters prepare
         params = {
@@ -1550,7 +1619,7 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/structures{}'.format(
+        path = 'structures{}'.format(
             f'/{extension}' if extension else ''
         )
 
@@ -1580,7 +1649,7 @@ class AsyncClient(AsyncOAuthClient):
             raise KeyError("Must specify an `applicationName`")
 
         # path prepare
-        path = '/WebApi/sync{}'.format(
+        path = 'sync{}'.format(
             f'/{extension}' if extension else ''
         )
 
@@ -1608,7 +1677,7 @@ class AsyncClient(AsyncOAuthClient):
     ) -> Union[Dict, List]:
         
         # path prepare
-        path = '/WebApi/sync{}'.format(
+        path = 'sync{}'.format(
             f'/{extension}' if extension else ''
         )
 
@@ -1645,4 +1714,4 @@ class AsyncClient(AsyncOAuthClient):
     async def get_version(self) -> Union[Dict, List]:
         
         # request.get -> json
-        return await self.get(path='/WebApi/version')
\ No newline at end of file
+        return await self.get(path='version')
\ No newline at end of file
